"""
implementation of the Box_dictionary class and its helper function
A Box_dictionary object contains the sketches present in each image. Its methods collect and modify the sketches
"""

from typing import Union,List,Dict,Tuple
from cryoloBM import constants,helper,MySketch,helper_filament
from cryolo import utils    # for type hinting purpose
from os import path
from math import isnan,sqrt as math_sqrt
from copy import deepcopy


class Box_dictionary:
    def __init__(self, type_case: constants.Type_case, onlyfiles: List[str], tot_slices: int = None  ) -> None:
        """
        It contains the sketches present in each image, SPA case, or slice, TOMO and TOMO_FOLDER cases.
        It is a dict of dicts. With the following structure:

            SPA case                    --> dict of list
        k1: micrograph's filename
        v1= list of sketches in the micrograph

            TOMO or TOMO_FOLDER cases   --> dict of dict of list
        k1: tomo filename
        v1= dict
            k2: number of slice of the 'k1' filename
            d2: list of sketches in the slice

        The functions 'get_sketches' and 'add_sketch' get and add sketches in the correct location

        :param type_case: Identify if the loaded case is SPA, single tomo or folder of tomo (Constant.Type_case)
        :param onlyfiles: list of the files present in the folder, single file (into a var list)in case of TOMO
        :param tot_slices: number of slices present in the TOMO. Only constants.Type_case.TOMO case
        """

        self.counter = 1    # used to assign an unique id to all the particles/filament present the dict. (it is a global counter and it counts in the tomo folder too)

        self.type_case = constants.Type_case.NOT_LOADED.value
        self.is_3D = False # True if the 3D visualization is available only in tomo cases after loading cbox

        self.is_cbox_untraced = False
        self.has_filament = False

        self.list_purefiles = list()    # list of file without extension

        # flag to detect the correct dictionary in case of filament. False select list of particles
        self.fil_start_end_vis = True

        # The following dictionary are linked by keys
        # When we change the image the boxes present in the old image will be deleted and the boxes present in the new one will be drawn
        # When we plot the new boxes we check if the boxsize or the visualization is changed from the last time they are drawn
        #   in case we transform them in real time, we update them in the particle_dictionary and the values in the other 2 dicts

        # the following dicts contain the boxes present in the images. The dict we ll use in the run depends to the context
        self.particle_dictionary = dict()       # 2D particle, it could represent the 'segmented' visualization of the filament case
        self.particle3D_dictionary = dict()     # 3D representation of 2D particle (referred to box_dict_traced in case of tracing) it could represent the 'segmented' visualization of the filament case
        self.box_dict_traced=dict()             # 2D representation of 2D untraced cbox sketches after running the tracing

        self.cryolo_filament_dictionary = dict()  # the filament got via cryolo in cryolo.utils.filament obj
        self.filament_dictionary = dict()       # 2D filament in rect_start_end visualization
        self.filament3D_dictionary = dict()     # 3D representation of 2D filament (referred to box_dict_traced in case of tracing)

        # it contains the boxsize of all the groups of sketches present in avery image and saved in particle_dictionary or filament dict
        # in filament case is referred to the base of the rect
        self.boxsize_dictionary = dict()
        # it contains the visualization of all the groups of sketches present in avery image and saved in particle_dictionary or filament dict
        self.visualization_dictionary = dict()
        # it contains the box distance between 2 sketches, which belong to a filament, only filament case
        self.box_distance_dictionary = dict()

        self.init(type_case=type_case,onlyfiles=onlyfiles, tot_slices=tot_slices)

    def init(self,type_case: constants.Type_case, onlyfiles: List[str], tot_slices: int) -> None:
        """
        It init the vars, we need it because after resetting we have to build the structure of the dicts from scratch.
        If we was working in TOMO and then we switch to TOMO_FOLDER we cannot reuse the old structure.
        Hence in the code we call clean and then in case we open a new folder/image we will call it manually again
        :param type_case: Identify if the loaded case is SPA, single tomo or folder of tomo (Constant.Type_case)
        :param onlyfiles: list of the files present in the folder, single file (into a var list)in case of TOMO
        :param tot_slices: number of slices present in the TOMO. Only constants.Type_case.TOMO case
        """
        self.list_purefiles = [path.splitext(path.basename(f))[0] for f in onlyfiles]
        self.type_case = type_case

        if self.type_case == constants.Type_case.TOMO.value:
            f=path.splitext(path.basename(onlyfiles[0]))[0]
            self.particle_dictionary.update({f: dict()})
            self.particle3D_dictionary.update({f: dict()})
            self.box_dict_traced.update({f: dict()})
            self.boxsize_dictionary.update({f: -1})
            self.visualization_dictionary.update({f: dict()})
            self.box_distance_dictionary.update({f: dict()})
            self.filament_dictionary.update({f: dict()})
            self.cryolo_filament_dictionary.update({f: dict()})
            self.filament3D_dictionary.update({f: dict()})
            self.init_single_tomo(tomo_name=f, tot_slices=tot_slices)
        else:
            self._init_dict_folder_case(onlyfiles=onlyfiles)


    def _init_dict_folder_case(self, onlyfiles:List[str])-> None:
        """
        Init all the dicts
        :param onlyfiles: list of filename, each filename identifies an image
        """
        for f in onlyfiles:
            self.particle_dictionary.update({path.splitext(path.basename(f))[0]: list() if self.type_case == constants.Type_case.SPA.value else dict()})
            self.particle3D_dictionary.update({path.splitext(path.basename(f))[0]: list() if self.type_case == constants.Type_case.SPA.value else dict()})
            self.boxsize_dictionary.update({path.splitext(path.basename(f))[0]: list() if self.type_case == constants.Type_case.SPA.value else dict()})
            self.visualization_dictionary.update({path.splitext(path.basename(f))[0]:  None  if self.type_case == constants.Type_case.SPA.value else dict()})
            self.box_distance_dictionary.update({path.splitext(path.basename(f))[0]: None if self.type_case == constants.Type_case.SPA.value else dict()})
            self.filament_dictionary.update({path.splitext(path.basename(f))[0]: list() if self.type_case == constants.Type_case.SPA.value else dict()})
            self.cryolo_filament_dictionary.update({path.splitext(path.basename(f))[0]: list() if self.type_case == constants.Type_case.SPA.value else dict()})
            self.filament3D_dictionary.update({path.splitext(path.basename(f))[0]: list() if self.type_case == constants.Type_case.SPA.value else dict()})
            if self.type_case != constants.Type_case.SPA.value:
                self.box_dict_traced.update({path.splitext(path.basename(f))[0]:  dict()})

    def init_single_tomo(self, tomo_name:str, tot_slices:int)-> None:
        """
        Init the dict of a generic tomo
        :param tomo_name: filenae tomo to init
        :param tot_slices: number of slices present in the TOMO.
        """
        if not self.particle_dictionary.get(tomo_name,dict()):
            self.particle_dictionary[tomo_name] = {sl: list() for sl in range(tot_slices)}
            self.particle3D_dictionary[tomo_name] = {sl: list() for sl in range(tot_slices)}
            self.box_dict_traced[tomo_name] = {sl: list() for sl in range(tot_slices)}
            self.boxsize_dictionary[tomo_name] = {sl: -1 for sl in range(tot_slices)}
            self.visualization_dictionary[tomo_name] = {sl: None for sl in range(tot_slices)}
            self.box_distance_dictionary[tomo_name] = {sl: None for sl in range(tot_slices)}
            self.filament_dictionary[tomo_name] = {sl: list() for sl in range(tot_slices)}
            self.cryolo_filament_dictionary[tomo_name] = {sl: list() for sl in range(tot_slices)}
            self.filament3D_dictionary[tomo_name] = {sl: list() for sl in range(tot_slices)}

    def get_sketches(self,f:str, n:int=None, is_3D:bool = False, is_tracing:bool = False) ->List[MySketch.MySketch]:
        """
        returns the sketches present in the given file 'f' at the given slice 'n' (in TOMO cases)
        :param f: filename
        :param n: number of slice
        :param is_3D if True return the 3D dictionary. Always False for updating counter purposes
        :param is_tracing: if True it search into the tracing dict (box_dict_traced) otherwise (particle_dictionary)
        :return: list of Mysketch obj present in the given file 'f' at the given slice 'n'
        """
        if self.has_filament and self.fil_start_end_vis:
            dict_2D = self.filament_dictionary
            dict_3D = self.filament3D_dictionary
        else:
            dict_2D = self.box_dict_traced if is_tracing else self.particle_dictionary
            dict_3D = self.particle3D_dictionary
        if self.type_case == constants.Type_case.SPA.value:
            return dict_2D.get(f,list())
        return dict_3D.get(f,list()).get(n,list()) if is_3D else dict_2D.get(f,list()).get(n,list())

    def get_particle_dictionary_as_bbox(self,f:str)->Dict[int,List[utils.BoundBox]]:
        """
        convert the tomo dict of 'f' into a dict of bboxes. it is used only i the tracing via cryolo.grouping3d.do_tracing
        """
        d=dict()
        for n, sketches in self.particle_dictionary[f].items():
            d.update({n: [s.get_as_BBox(confidence=s.get_confidence(),z=n) for s in sketches]})
        return d

    def resize_3D(self,new_boxsize:int, old_boxsize:int, use_estimated_size:bool = False)-> None:
        """
        Reset the particle3D_dictionary, copy the value in particle_dictionary, resize them and create3D for each particle
        In the process it resizes the whole 2D dict too
        :param new_boxsize: the new boxsize value
        :param old_boxsize: the old boxsize value. used for filament 3D for creating the ratio value
        :param use_estimated_size: if True uses the estimated value present in the box (Mysketch obj)
        """
        for f in self.particle_dictionary.keys():
            # reset the 3D dict only in particle case
            for n in self.particle_dictionary[f].keys():
                self.boxsize_dictionary[f][n] =new_boxsize
                if not self.has_filament:
                    self.particle3D_dictionary[f][n] =list()
            # fill the 3D dict and resize the 2D dict
            part_dict = self.particle3D_dictionary[f] if self.has_filament else self.particle_dictionary[f]
            for n,sketches in part_dict.items():
                for sketch in sketches:
                    if self.has_filament:
                        # in this case the "estiamte option is blurred so i do not have to take it in consideration
                        sketch.resize(sketch.get_width()*(new_boxsize/old_boxsize))
                    else:
                        sketch.resize(sketch.get_est_size(circle=False) if use_estimated_size else new_boxsize)	        # resize the sketch of particle_dictionary
                        self.create_3D(sketch,f,n, is_start_end_filament = False)

    def _resize_all_current_filament(self, new_boxsize:int, f:str, n:int=None,is_saving_on_file=False)->List[MySketch.MySketch]:
        """
        Resize and return the sketches in the present in the given file 'f'
        ONLY 2D cases
        :param new_boxsize: the new boxsize value
        :param f: filename
        :param n: number of slice. The untraced cbox are 2D but on Tomo
        :param is_saving_on_file: True if we are saving on file
        :return: list of resized Mysketch obj present in the given file 'f'
        """
        starting_vis = self.has_same_visualization(visualization=constants.Visualization_cb.RECT_FILAMENT_START_END.value,f=f,n=n)

        # set temporarily to True for getting via 'get_sketches' the start_end rect
        self.fil_start_end_vis = True
        fil_start_end = self.get_sketches(f=f, n=n, is_3D=self.is_3D, is_tracing=False)
        self.fil_start_end_vis = False
        fil_segmented = self.get_sketches(f=f, n=n, is_3D=self.is_3D, is_tracing=False)
        self.fil_start_end_vis = starting_vis

        # vars to track colour and list of sketches of a single filament
        listed_sketches = {fil.id: list() for fil in fil_start_end}
        colour_sketches = {fil.id: fil.get_color() for fil in fil_start_end}

        # resize the segmented visualization
        for box in fil_segmented:
            box.resize(new_size=new_boxsize)
            listed_sketches[box.id].append(box.get_xy(circle=False))

        # resize, recrating and overwriting on self.dictionary the start-end visualization
        fil_start_end = list()
        for fil_id, xy in listed_sketches.items():
            fil = helper_filament.picked_filament(box_size=new_boxsize, is_tomo=self.type_case != constants.Type_case.SPA.value, fil_id=fil_id)
            fil.begin_fil = listed_sketches[fil_id][0][0] +new_boxsize/2, listed_sketches[fil_id][0][1] +new_boxsize/2
            fil.end_fil = listed_sketches[fil_id][-1][0] +new_boxsize/2,  listed_sketches[fil_id][-1][1] +new_boxsize/2
            fil_start_end.append(fil.get_rect_sketch(color=colour_sketches[fil_id], sketch_id=fil_id))

        # i have to set the new filament in start-end visualization otherwise they ll be not saved in filemant2D[the current image]
        self.fil_start_end_vis = True
        self._set_sketches(list_sketches=fil_start_end, f=f, n=n)
        self.fil_start_end_vis = starting_vis
        if is_saving_on_file:
            return fil_segmented
        return fil_start_end if starting_vis else fil_segmented


    def resized_and_get_sketches(self, new_boxsize:int, f:str, n:int=None, use_estimated_size:bool = False,is_saving_on_file=False)->List[MySketch.MySketch]:
        """
        Resize and return the sketches in the present in the given file 'f'
        ONLY 2D cases
        :param new_boxsize: the new boxsize value
        :param f: filename
        :param n: number of slice. The untraced cbox are 2D but on Tomo
        :param use_estimated_size: if True uses the estimated value present in the box (Mysketch obj)
        :param is_saving_on_file: True if we are saving on file
        :return: list of resized Mysketch obj present in the given file 'f'
        """
        if self.has_filament:
            boxes=self._resize_all_current_filament(new_boxsize=new_boxsize, f=f, n=n,is_saving_on_file=is_saving_on_file)
        else:
            boxes = self.get_sketches(f=f,n=n, is_3D=self.is_3D,  is_tracing = False)
            for box in boxes:
                box_new_size = box.get_est_size(circle=False) if use_estimated_size else new_boxsize
                box.resize(new_size=box_new_size)

        # update the boxsize dictionary, i force it to be None in order to recalculate the boxes next time
        self.update_boxsize_dictionary(new_boxsize=None if use_estimated_size else new_boxsize, f=f, n=n)
        return boxes

    def _set_sketches(self,list_sketches: List[MySketch.MySketch],f:str, n:int=None)->None:
        """
        After changing the boxes inthe list, if it was done via for box in boxes i have to overwrite the list
        ONLY 2D cases
        :param list_sketches: the new boxsize value
        :param f: filename
        :param n: number of slice. The untraced cbox are 2D but on Tomo
        """

        dict_2D = self.filament_dictionary if self.has_filament and self.fil_start_end_vis else self.particle_dictionary
        if self.type_case == constants.Type_case.SPA.value:
            dict_2D[f] = list_sketches
        else:
            dict_2D[f][n] = list_sketches

    def change_box_size_all(self, new_box_size:int,is_saving_on_file:bool=True)->None:
        """
        change the distance between the sketches in all the dict, it is used before saving on file
        :param new_box_size: new boxsize of a sketch (of course it is referred to the segmented mode)
        :param is_saving_on_file: True if we are saving on file
        """
        starting_vis = self.fil_start_end_vis
        self.fil_start_end_vis = False
        for k1 in self.particle_dictionary.keys():
            if self.type_case == constants.Type_case.SPA.value:
                self.particle_dictionary[k1] = self.resized_and_get_sketches(new_boxsize=new_box_size, f=k1, n=None, use_estimated_size=False,is_saving_on_file=is_saving_on_file)
            else:
                for k2 in self.particle_dictionary[k1].keys():
                    self.particle_dictionary[k1][k2] = self.resized_and_get_sketches(new_boxsize=new_box_size, f=k1, n=k2,  use_estimated_size=False,is_saving_on_file=is_saving_on_file)
        self.fil_start_end_vis = starting_vis

    def change_box_distance_all(self, new_box_distance:int)->None:
        """
        change the distance between the sketches in all the dict, it is used before saving on file
        :param new_box_distance: the new distance between 2 sketches in a filament
        """
        starting_vis = self.fil_start_end_vis
        self.fil_start_end_vis = False
        for k1 in self.particle_dictionary.keys():
            if self.type_case == constants.Type_case.SPA.value:
                self.particle_dictionary[k1] = self.change_box_distance(new_box_distance=new_box_distance, f=k1, n=None, is_tracing=False)
            else:
                for k2 in self.particle_dictionary[k1].keys():
                    self.particle_dictionary[k1][k2] = self.change_box_distance(new_box_distance=new_box_distance, f=k1, n=k2,  is_tracing=False)
        self.fil_start_end_vis = starting_vis

    def set_colors_all(self, new_color:str)->None:
        """
        Set the same color for all the filaments. Used to compare the TRACED results. (called in multiple import)
        :param new_color: new color
        """
        def _set_colors(particles:List[MySketch.MySketch], filaments:List[MySketch.MySketch])->None:
            """
            Set the colors of a list of Mysketch obj present in the given lists
            :param particles: list of Mysketches obj
            :param particles: list of Mysketches obj
            """
            for p in particles:
                p.set_colour(colour=new_color)
            for f in filaments:
                f.set_colour(colour=new_color)

        dict_fil = self.filament3D_dictionary if self.is_3D else self.filament_dictionary
        dict_part = self.particle3D_dictionary if self.is_3D else self.particle_dictionary
        if self.type_case == constants.Type_case.SPA.value:
            for micrograph in dict_part.keys():
                _set_colors(particles=dict_part[micrograph], filaments=dict_fil[micrograph])
        else:
            for tomo_name in dict_part.keys():
                for z in dict_part[tomo_name].keys():
                    _set_colors(particles=dict_part[tomo_name][z], filaments=dict_fil[tomo_name][z])

    def get_boxsize(self, f:str, n:int=None)->int:
        """
        get the boxsize of a given file (and slice)
        :param f: filename
        :param n: number of slice
        :return boxsize
        """
        return self.boxsize_dictionary.get(f,-1) if self.type_case == constants.Type_case.SPA.value else self.boxsize_dictionary.get(f,list()).get(n,-1)


    def change_box_distance(self, new_box_distance:int, f:str, n:int=None, is_tracing:bool = False)-> List[MySketch.MySketch]:
        """
        change the distance between the sketches in the present in the given file 'f' at the given slice 'n' (in TOMO cases)
        :param new_box_distance: the new distance between 2 sketches in a filament
        :param f: filename
        :param n: number of slice
        :param is_tracing: if True it search into the tracing dict (box_dict_traced) otherwise (particle_dictionary)
        :return new_list
        """
        new_list = list()

        fil_segmented = self.get_sketches(f=f, n=n, is_3D=self.is_3D, is_tracing=is_tracing)
        listed_sketches = {fil.id: list() for fil in fil_segmented}
        colour_sketches = {fil.id: fil.get_color() for fil in fil_segmented}

        for box in fil_segmented:
            if box.id in listed_sketches:
                z = int(box.z) if box.z is not None and not isnan(box.z) else box.z  # otherwise if z is a float it crashes in utils.resample_filament
                listed_sketches[box.id].append(box.get_as_BBox(confidence=box.get_confidence(), z=z))

        # Create the new list of boxes
        for fil_id, sketches in listed_sketches.items():
            fil = utils.Filament(boxes=sketches)
            resampled_f=utils.resample_filament(filament=fil,new_distance=new_box_distance)
            new_list.extend(helper.convert_list_bbox_to_sketch(boxes_list=resampled_f.boxes,out_dict=False,col=colour_sketches[fil_id],fil_id=fil_id))

        self._set_sketches(list_sketches=new_list,f=f, n=n)
        return new_list


    def update_boxsize_dictionary(self, new_boxsize:int, f:str,n:int=None)-> None:
        """
        Resize and return the sketches in the present in the given file 'f' at the given slice 'n' (in TOMO cases)
        :param new_boxsize: the new boxsize value
        :param f: filename
        :param n: number of slice
        """

        if self.type_case == constants.Type_case.SPA.value:
            self.boxsize_dictionary[f] = new_boxsize
        else:
            self.boxsize_dictionary[f][n] = new_boxsize

    def update_visualization_dictionary(self, new_visualization:constants.Visualization_cb, f:str, n:int=None)-> None:
        """
        Resize and return the sketches in the present in the given file 'f' at the given slice 'n' (in TOMO cases)
        :param new_visualization: the new visualization value
        :param f: filename
        :param n: number of slice
        :return: list of resized Mysketch obj present in the given file 'f' at the given slice 'n'
        """
        if self.type_case == constants.Type_case.SPA.value:
            self.visualization_dictionary[f] = new_visualization
        else:
            self.visualization_dictionary[f][n] = new_visualization

    def update_box_distance_dictionary(self, new_box_distance:int, f:str, n:int=None)-> None:
        """
        Resize and return the sketches in the present in the given file 'f' at the given slice 'n' (in TOMO cases)
        :param new_box_distance: the new distance between 2 sketches in a filament
        :param f: filename
        :param n: number of slice
        """
        if self.type_case == constants.Type_case.SPA.value:
            self.box_distance_dictionary[f] = new_box_distance
        else:
            self.box_distance_dictionary[f][n] = new_box_distance

    def delete_sketch_by_index(self, index:int, f:str, n:int=None)-> int:
        """
        remove the index sketch in the given self.boxmanager[f][n] (in case of TOMO or TOMO_FOLDER case)
        :param index: index in the dicts of Mysketch obj to remove
        :param f: filename
        :param n: number of slice
        :return mysketch id. We need it when we call this function from 'delete_filemant_by_index'
        """
        dict_2D = self.filament_dictionary if self.has_filament else self.particle_dictionary
        if self.type_case == constants.Type_case.SPA.value:
            id_part = dict_2D[f][index].id
            del dict_2D[f][index]
        else:
            id_part = dict_2D[f][n][index].id
            del dict_2D[f][n][index]
            if self.is_3D:  # the untraced cbox sketches are into a TOMO but 2D
                self.delete_3D(id_part, f, n)
        return id_part

    def delete_filemant_by_index(self, index:int, f:str, n:int=None)-> None:
        """
        remove all the sketches, rect and not, present in all the dicts form the given 'fil_id' id
        :param index: index in the dicts of Mysketch filament to remove
        :param f: filename
        :param n: number of slice
        """
        id_part = self.delete_sketch_by_index(index=index,f=f,n=n)
        d = self.particle_dictionary[f] if self.type_case == constants.Type_case.SPA.value else self.particle_dictionary[f][n]
        for i in range(len(d) - 1, -1, -1):
            if d[i].id == id_part:
                del d[i]
        if self.is_3D:
            pass #how manage it? directly in self.delete_3D?

    def get_central_n(self, id_part:int, f:str, n:int)->Union[Tuple[MySketch.MySketch,int],Tuple[None,None]]:
        """
        find the central 2D sketch of the given 'id_part' particle
        :param id_part: id particle
        :param f: filename
        :param n: number of slice that represent the center
        :return the central box and its slice's number, None in case of error
        """
        tot_slice = len(self.particle_dictionary[f])
        real_boxsize = self.get_boxsize(f=f,n=n)
        box_size = real_boxsize if real_boxsize>0 else tot_slice
        prev_slices = range(n, max(n-box_size , 0), -1)
        next_slices = range(n, min(n + box_size , tot_slice - 1))
        for slice_n in list(prev_slices) + list(next_slices):
            for b in self.get_sketches(f=f, n=slice_n, is_3D=True,  is_tracing = False):
                if b.id == id_part and not b.only_3D_visualization:
                    return b,slice_n
        return None,None


    def delete_3D(self, id_part:int, f:str, n:int)-> None:
        """
        remove the sketches that belong to the 'id_part' particle
        :param id_part: id particle
        :param f: filename
        :param n: number of slice that represent the center
        """
        tot_slice = len(self.particle_dictionary[f])
        real_boxsize = self.get_boxsize(f=f, n=n)
        box_size = real_boxsize if real_boxsize > 0 else tot_slice
        prev_slices = range(n, max(int(n - box_size / 2), 0), -1)
        next_slices = range(n, min(int(n + box_size / 2), tot_slice - 1))
        for slice_n in list(prev_slices)+list(next_slices):
            ref_list = self.get_sketches(f=f, n=slice_n, is_3D=True,  is_tracing = False)
            for b in ref_list:
                if b.id == id_part:
                    del ref_list[ref_list.index(b)]
                    break

    def add_picked_filament(self, fil:MySketch.MySketch,f:str, n:int=None, still_picking:bool= True, fil_id=None)-> int:
        """
        insert the given sketch in the given self.boxmanager[f][n] (in case of TOMO or TOMO_FOLDER case)
        :param fil: instance of Mysketch obj
        :param f: filename
        :param n: number of slice
        :param still_picking: True if we are still picking
        :param fil_id: in case we resize the filament in 3D case (hence after collecting via cryolo we have to keep the their id
        :return filament id, its list of sketches has to have the same id (each sketch of this list). I need it in the import from file
        """
        # since for the real time visualization in the picking phase we delete and replot the current filament its id has to be always the same
        fil.id = fil_id if fil_id else self.counter

        # we update the counter only when we start to pick (onclick function). The first has counter 0 (in this way we do not have to adapt the code after importing)
        if not still_picking:
            self.counter+=1
        if self.type_case == constants.Type_case.SPA.value:
            self.filament_dictionary[f].append(fil)
        else:
            self.filament_dictionary[f][n].append(fil)
            if self.is_3D:  # the untraced cbox sketches are into a TOMO but 2D
                self.create_3D(fil,f,n, is_start_end_filament = True)
        return fil.id



    def add_sketch(self,sketch:MySketch.MySketch, f:str, n:int=None)-> None:
        """
        insert the given sketch in the given self.boxmanager[f][n] (in case of TOMO or TOMO_FOLDER case)
        :param sketch: instance of Mysketch obj
        :param f: filename
        :param n: number of slice
        """
        # first particle id is 0
        sketch.id = self.counter
        self.counter+=1
        if self.type_case == constants.Type_case.SPA.value:
            self.particle_dictionary[f].append(sketch)
        else:
            self.particle_dictionary[f][n].append(sketch)
            if self.is_3D:  # the untraced cbox sketches are into a TOMO but 2D
                self.create_3D(sketch,f,n, is_start_end_filament = False)

    def add_sketches(self,sketches:List[MySketch.MySketch], f:str, n:int=None, id_fil = None, colour:str = 'r')-> None:
        """
        It basically use in the importing data step
        insert the given sketches in the given self.boxmanager[f][n] (in case of TOMO or TOMO_FOLDER case)
        :param sketches: list of instances of Mysketch obj
        :param f: filename
        :param n: number of slice
        :param id_fil: id particles (they have the same id if they belong to a filament)
        :param colour: colour of the sketched. In filamnet they have the same colour of the filament
        """
        # first particle id is 0
        for s in sketches:
            s.set_colour(colour=colour)
            if id_fil is not None:
                s.id = id_fil
            else:
                s.id=self.counter
                self.counter+=1
        if self.type_case == constants.Type_case.SPA.value:
            self.particle_dictionary[f].extend(sketches)
        else:
            self.particle_dictionary[f][n].extend(sketches)
            if self.is_3D:  # the untraced cbox sketches are into a TOMO but 2D
                for s in sketches:
                    self.create_3D(sketch=s,f=f,n=n, is_start_end_filament = False)

    def create_3D(self, sketch: MySketch.MySketch, f: str, n: int, is_start_end_filament: bool) -> None:
        """
        It basically use in the importing data step
        insert the 2D section of sketch in the given self.boxmanager[f][n] (in case of TOMO or TOMO_FOLDER case)
        for representing the 3D visualization of it
        :param sketch: instance of Mysketch obj of which we create the 3D visualization
        :param f: filename
        :param n: number of slice
        :param is_start_end_filament: True if it is creating a 3D for start end filament visualization
        """

        # added the given sketch to the 3D dict
        dict3D = self.filament3D_dictionary if is_start_end_filament else self.particle3D_dictionary
        dict3D[f][n].append(sketch)

        tot_slice = len(self.filament_dictionary[f]) if is_start_end_filament else len(self.particle_dictionary[f])
        original_size = int(sketch.get_height())
        # i do not care if len(prev_slices)==len(next_slices) because 'zip' ll consider the short one length as limit
        # in real case this situation should not happened because a particle is a sphere and the whole particle
        # is into the tomo.
        # Hence cryolo will not lead us to this situation and the user ll not do either
        prev_slices = range(n - 1, max(int(n - original_size / 2), 0), -1)  # /2 because new_size-=2
        next_slices = range(n + 1, min(int(n + original_size / 2), tot_slice - 1))
        new_size = original_size - 2
        delta = 1

        for prev_index, next_index in zip(prev_slices, next_slices):
            new_sketch = helper.create_sketch(sketch=sketch, delta=delta, size=new_size,
                                              is_start_end_filament=is_start_end_filament)
            new_sketch.z = prev_index
            dict3D[f][prev_index].append(new_sketch)

            new_sketch2 = deepcopy(new_sketch)  # i have to deepcopy because they otherwise they are the same obj (hence in both the slices i'd have the same z)
            new_sketch2.z = next_index
            dict3D[f][next_index].append(new_sketch2)
            new_size -= 2
            delta += 1

    def create_traced_3D(self)-> None:
        """
        Create the 3D results after running the tracing.
        The tracing result are into 'box_dict_traced'. I create the 3D version of these values
        """
        for f in self.box_dict_traced.keys():
            for n,sketches in self.box_dict_traced[f].items():
                for sketch in sketches:
                    self.create_3D(sketch=sketch,f=f,n=n,is_start_end_filament=False)

    @staticmethod
    def get_upper_lower_thres_tomo(single_box_dict:dict)->Tuple[int,int]:
        """
        The 'minimum size' and 'maximum size' of the thresholding tab.
        :param single_box_dict: box dict of a case (SPA o image TOMO)
        """
        min_size = 99999
        max_size = -99999
        for _, rectangles in single_box_dict.items():
            for rect in rectangles:
                if rect.get_est_size(circle=False) > max_size:
                    max_size = rect.get_est_size(circle=False)
                if rect.get_est_size() < min_size:
                    min_size = rect.get_est_size(circle=False)


        #in case you selected an empty folder you have to switch them
        if max_size<min_size:
            max_size,min_size=min_size,max_size

        return max_size, min_size

    def get_upper_lower_thres(self)-> Union[Tuple[int,int],Dict[str,Dict[str,int]]]:
        """
        After loading from cbox we have to set dynamically the lower and upper threshold.
        The 'minimum size' and 'maximum size' of the thresholding tab.
        In case of tomo, both the cases, return a dict
            k: filename
            v: [upper value, lower value]
        :return upper and lower values
        """

        if self.type_case == constants.Type_case.SPA.value:
            return self.get_upper_lower_thres_tomo(single_box_dict=self.particle_dictionary)
        else:
            out = dict()
            for filename in self.particle_dictionary.keys():
                mx,mn =self.get_upper_lower_thres_tomo(single_box_dict=self.particle_dictionary[filename])
                out.update({ filename:{ "upper":mx,"lower":mn}})
            return out

    def has_same_boxsize(self,boxsize:int, f:str, n:int=None)->bool:
        """
        Check if in the file 'f' slice 'n' (only tomo cases) last time it was plotted with the same boxsize
        :param boxsize: boxsize value
        :param f: filename
        :param n: number of slice
        :return: true if a resize is not needed
        """
        return boxsize == self.get_boxsize(f=f,n=n)

    def has_same_visualization(self, visualization:constants.Visualization_cb, f:str, n:int=None)->bool:
        """
        Check if in the file 'f' slice 'n' (only tomo cases) last time it was plotted with the same visualization
        :param visualization: visualization value
        :param f: filename
        :param n: number of slice
        :return: true if a resize is not needed
        """
        if self.type_case == constants.Type_case.SPA.value:
            return visualization == self.visualization_dictionary.get(f)
        return visualization == self.visualization_dictionary.get(f).get(n)


    def has_same_box_distance(self, box_distance:int, f:str, n:int=None)->bool:
        """
        Check if in the file 'f' slice 'n' (only tomo cases) last time it was plotted with the same box distance
        :param box_distance: box_distance value
        :param f: filename
        :param n: number of slice
        :return: true if a resize is not needed
        """
        if self.type_case == constants.Type_case.SPA.value:
            return box_distance == self.box_distance_dictionary.get(f)
        return box_distance == self.box_distance_dictionary.get(f).get(n)

    def is_empty(self)->bool:
        """
        Return True if there are no boxes in the particle_dictionary var
        """
        if self.type_case == constants.Type_case.SPA.value:
            for f in self.particle_dictionary.keys():
                if len(self.particle_dictionary[f]) > 0:
                    return False
        else:
            for f in self.particle_dictionary.keys():
                for n in self.particle_dictionary[f].keys():
                    if len (self.particle_dictionary[f][n]) > 0:
                        return False
        return True

    def update_counter(self, n=1):
        """
        Update the counter
        :param n: number of unit to add
        """
        self.counter+=n

    def clean(self)->None:
        """
        clean the dicts
        """
        self.particle_dictionary = dict()
        self.particle3D_dictionary = dict()
        self.filament_dictionary = dict()
        self.cryolo_filament_dictionary = dict()
        self.filament3D_dictionary = dict()
        self.boxsize_dictionary = dict()
        self.visualization_dictionary = dict()
        self.box_distance_dictionary = dict()
        self.box_dict_traced = dict()
        self.type_case=constants.Type_case.NOT_LOADED.value
        self.is_3D = False
        self.counter = 1
        self.is_cbox_untraced = False
        self.list_purefiles = list()

    def set_untraced_param(self)->None:
        """
        In the TOMO when we load untraced values we are loading 2D particles
        """
        self.is_3D = False
        self.is_cbox_untraced = True

    def get_tot_filament(self,tomo_name=None)->int:
        """
        Get the total number of filament
        :param tomo_name: tomo fname, in case tomofolder with None name it ll count all the written filaments
        """
        tot = 0
        if self.type_case == constants.Type_case.SPA.value:
            for f in self.filament_dictionary.keys():
                tot+=len(self.filament_dictionary[f])
        else:
            if tomo_name is not None:
                for z in self.filament_dictionary[tomo_name].keys():
                    tot += len(self.filament_dictionary[tomo_name][z])
            else:
                for f in self.filament_dictionary.keys():
                    tot+=self.get_tot_filament(f)
        return tot

    def set_box_distance_dict_after_import(self):
        """
        After loading from file it has to calculate the box distance and set it
        """
        def calculate_dist():
            if len(sketches) > 3:
                x1, y1 = sketches[0].get_xy(circle=True)
                for sketch in sketches:
                    x2, y2 = sketch.get_xy(circle=True)
                    if x1 != x2 or y1 != y2:
                        dist = int(math_sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2))
                        return dist

        d=None
        if self.type_case == constants.Type_case.SPA.value:
            for sketches in self.particle_dictionary.values():
                d= calculate_dist()
                if d is not None:
                    for k in self.box_distance_dictionary.keys():
                        self.box_distance_dictionary[k]=d
                    return d
        else:
            for tomo_name in self.particle_dictionary.keys():
                for sketches in self.particle_dictionary[tomo_name].values():
                    d = calculate_dist()
                    if d is not None:
                        for tname in self.box_distance_dictionary.keys():
                            for z in self.box_distance_dictionary[tname].keys():
                                self.box_distance_dictionary[tname][z] = d
                        return d
        return d

    def reset(self)->None:
        """
        Remove all the sketches present into self.particle_dictionary.
        It is called after reset->file
        """
        self.counter = 1
        self.is_cbox_untraced = False
        self.has_filament = False

        for k1,v1 in self.particle_dictionary.items():
            if self.type_case == constants.Type_case.SPA.value:
                self.particle_dictionary[k1] = list()
                self.boxsize_dictionary[k1] = -1
                self.particle3D_dictionary[k1] = list()
                self.filament_dictionary[k1] = list()
                self.cryolo_filament_dictionary[k1] = list()
                self.filament3D_dictionary[k1] = list()
                self.visualization_dictionary[k1] = None
                self.box_distance_dictionary[k1] = None
            else:
                for k2,v2 in self.particle_dictionary[k1].items():
                    self.particle_dictionary[k1][k2] = list()
                    self.boxsize_dictionary[k1][k2] = -1
                    self.particle3D_dictionary[k1][k2] = list()
                    self.filament_dictionary[k1][k2] = list()
                    self.cryolo_filament_dictionary[k1][k2] = list()
                    self.filament3D_dictionary[k1][k2] = list()
                    self.visualization_dictionary[k1][k2] = None
                    self.box_distance_dictionary[k1][k2] = None




"""
In this file are implemented all the function used for interacting with the image in the GUI, for istance:
-) 'onclick' when you click on the image
-) 'onrelease' when you release the button after clicking on the image
-) 'box_size_changed' when the boxsize is changed
-) 'remove_picked_filament' for removing a filament
....
....
"""


from cryoloBM import constants,helper, helper_GUI, MySketch,visualization_3D_Window, helper_tab_tracing, helper_filament
from cryolo import imagereader
from os import path
import numpy as np

from cryoloBM import boxmanager_controller # for type hinting purpose
from typing import Tuple

# EVENT_IMAGE_CHANGED FUNCTIONS
def _check_event_image_changed_validity(controller:boxmanager_controller, root_tree_item:helper_GUI.QtG.QTreeWidgetItem)->bool:
    """
    It is for internal use of 'event_image_changed'. Check if we can change the image
    :param controller: Boxmanager_controller obj
    :param root_tree_item: It is filled in automatic via PyQT
    Return False if we cannot change the image
    """
    if not root_tree_item or controller.model.type_case == constants.Type_case.NOT_LOADED.value:
        return False

    cond_true = [ controller.view.current_image_path, root_tree_item.childCount() == 0,
                 controller.model.type_case != constants.Type_case.NOT_LOADED.value]

    if not all(cond_true):
        return False
    return True

def _get_SPAimg_image_changed(controller:boxmanager_controller, old_filename:str)->Tuple[bool,np.ndarray]:
    """
    It is for internal use of 'event_image_changed'. Check if the new image has a different size
    it could happen that in a SPA case the micrographs have different size
    :param controller: Boxmanager_controller obj
    :return new_size_img: True if the size
    """
    prev_size = imagereader.read_width_height(image_path=path.join(controller.model.image_folder_path,old_filename))
    img = helper.read_image(file_path=controller.model.current_image_path, use_mmap=False)

    if prev_size[::-1] == img.shape:
        return False, img
    return True, img

def _set_and_update_boxdictionary_image_changed(controller:boxmanager_controller)->None:
    """
    When we open an image:
        -) the Boxdictionary's dictionaries has to be set
        -) the controller.model.rectangles has to be synchronized to the appropriate boxdictionary.boxdictionary entry
    :param controller: Boxmanager_controller obj
    """
    current_filename = controller.get_current_filename(with_extension=False)

    # if it is the first time we visualize the image we have to init boxsize_dictionary and visualization_dictionary
    if controller.model.box_dictionary.has_same_boxsize(boxsize=None, f=current_filename, n=controller.model.index_tomo):
        controller.model.box_dictionary.update_boxsize_dictionary(new_boxsize=controller.model.boxsize,
                                                                  f=current_filename, n=controller.model.index_tomo)
        controller.model.box_dictionary.update_visualization_dictionary(new_visualization=controller.model.visualization,
                                                                        f=current_filename, n=controller.model.index_tomo)

    # update the visualization
    if not controller.model.box_dictionary.has_same_visualization(visualization=controller.model.visualization,
                                                                  f=current_filename, n=controller.model.index_tomo):
        controller.model.box_dictionary.update_visualization_dictionary(
            new_visualization=controller.model.visualization,f=current_filename, n=controller.model.index_tomo)

    # load the sketches present in the current image/slice. If i need i resize them
    # NB the untraced cbox sketches are treat as 2D particle
    is_untraced = controller.model.is_cbox_untraced
    has_same_boxsize = controller.model.box_dictionary.has_same_boxsize(boxsize=controller.model.boxsize, f=current_filename, n=controller.model.index_tomo)

    if (not is_untraced and has_same_boxsize)\
            or(controller.model.has_filament and controller.model.has_3d):

        controller.model.rectangles = controller.model.box_dictionary.get_sketches(f=current_filename,
                                                        n=controller.model.index_tomo, is_3D =True,  is_tracing = False)
    else:
        controller.model.rectangles = controller.model.box_dictionary.resized_and_get_sketches(
            new_boxsize=controller.model.boxsize, f=current_filename, n=controller.model.index_tomo,use_estimated_size=controller.model.use_estimated_size)

    # When we load the filament, their particles have different size hence we do not want to lose these info and we do not allow to change the box distamce
    if controller.model.visualization in [constants.Visualization_cb.CIRCLE_SEGMENTED.value,constants.Visualization_cb.RECT_FILAMENT_SEGMENTED.value] and \
            (not controller.model.box_dictionary.has_same_box_distance(box_distance=controller.model.box_distance_filament, f=current_filename, n=controller.model.index_tomo) or controller.model.has_3d):
        controller.model.rectangles =controller.model.box_dictionary.change_box_distance(new_box_distance=controller.model.box_distance_filament, f=current_filename, n=controller.model.index_tomo)
    return

def event_image_changed(controller:boxmanager_controller, root_tree_item:helper_GUI.QtG.QTreeWidgetItem)->None:
    """
    Manage the 'image changes' event. It is called when we click on a micrograph's filename (or slice of a tomo)
    in the main window
    Change the displayed image and set all the internal vars
    :param controller: Boxmanager_controller obj
    :param root_tree_item: It is filled in automatic via PyQT
    """
    not_the_same_tomo = False  # var used to close the preview win (trace option) in tomo folder case
    if not _check_event_image_changed_validity(controller=controller, root_tree_item=root_tree_item):
        return

    # i do not need to call h_keypress() but only reset its related flag
    if controller.model.toggle:
        controller.model.toggle = False

    controller.view.current_tree_item = root_tree_item
    previous_filename = controller.get_current_filename(with_extension = True)

    # delete the patches present in the previous image/slice and load the patches present in the current one
    # they are already present in controller.model.rectangle
    if controller.model.box_dictionary:
        delete_all_patches(controller=controller)
        # set vars to identify current image/slice
        if controller.model.type_case == constants.Type_case.SPA.value:
            helper.set_current_image_path(controller=controller, current_image_path=path.join(controller.model.image_folder_path,
                                                    str(controller.view.current_tree_item.text(0))))
        else:
            if controller.model.type_case == constants.Type_case.TOMO_FOLDER.value:
                helper.set_current_image_path(controller=controller,
                                              current_image_path= path.join(controller.model.image_folder_path,root_tree_item.parent().text(0)))
            # set tomo vars
            controller.model.index_tomo = int(controller.view.current_tree_item.text(0))
            controller.model.last_index_tomo = controller.model.index_tomo

        # update the boxdictionary's dictionaries and set controller.model.rectangles
        _set_and_update_boxdictionary_image_changed(controller=controller)

    # it could happen that in a SPA case the micrographs have different size
    new_size_img = False

    if controller.model.type_case == constants.Type_case.SPA.value:
        title = controller.get_current_filename(with_extension = True)

        new_size_img,img = _get_SPAimg_image_changed(controller=controller, old_filename=previous_filename)

    else:
        if controller.model.type_case == constants.Type_case.TOMO.value:
            pass
        else:
            if previous_filename != controller.get_current_filename(with_extension = True):
                not_the_same_tomo = True
                pure_filename =controller.get_current_filename(with_extension = False)
                controller.view.memory_slider.setMaximum(controller.model.tot_frames[pure_filename] - 1)
                controller.view.min_length_slider.setMaximum(controller.model.tot_frames[pure_filename] - 1)
                controller.view.search_range_slider.setMaximum(controller.model.smallest_image_dim[pure_filename] - 1)
                oldsize_single_slice=controller.model.current_tomoimage_mmap.shape[1:3]
                controller.model.current_tomoimage_mmap = helper.read_image(file_path=controller.model.current_image_path, use_mmap=True)
                new_size_img = controller.model.current_tomoimage_mmap.shape[1:3] != oldsize_single_slice
                helper.load_tomo_params(controller=controller,pure_filename=pure_filename)

        img = controller.model.current_tomoimage_mmap[controller.model.index_tomo, :, :]
        title = f"{controller.get_current_filename(with_extension=True)} \tslice: {controller.model.index_tomo}"

    # set image properties and show it
    if new_size_img:
        controller.view.im = controller.view.ax.imshow(img, origin="lower", cmap="gray", interpolation="Hanning")
    else:
        controller.view.im.set_data(img)
    controller.view.setWindowTitle_viewPlot(title=title)
    controller.view.fig.canvas.draw()
    controller.view.background_current = controller.view.fig.canvas.copy_from_bbox(controller.view.ax.bbox)
    controller.view.background_current_onmove = controller.view.background_current

    # draw the sketches present in the current image/slice
    if controller.model.rectangles:
        draw_all_patches(controller=controller)

    if controller.view.preview_win and controller.view.preview_win.is_visible():
        if not_the_same_tomo:
            controller.view.preview_win.plot.close()
            if controller.model.box_dictionary.box_dict_traced[controller.get_current_filename(with_extension = True)]:
                controller.view.preview_win = visualization_3D_Window.PreviewWindow(controller.view.font, helper_tab_tracing.get_patches(controller))
                controller.view.preview_win.display_img(controller.get_current_filename(with_extension = False), img, controller.model.index_tomo)
        else:
            controller.view.preview_win.set_list_patches(helper_tab_tracing.get_patches(controller))
            controller.view.preview_win.display_img(controller.get_current_filename(with_extension = False), img, controller.model.index_tomo)



# ONCLICK FUNCTIONS
def _check_onclick_validity(controller:boxmanager_controller, event:helper_GUI.plt_qtbackend)->bool:
    """
    It is for internal use of 'onclick'. Check if we can interact, drawing or deleting sketch, to the image
    :param controller: Boxmanager_controller obj
    :param event: the click event
    Return False if we cannot interact with the image
    """

    if controller.model.has_filament and controller.model.has_3d:
        # We loaded traced filaments
        controller.view.err_message(msg="Picking is disabled.")
        return False

    if controller.model.has_filament and controller.model.visualization != constants.Visualization_cb.RECT_FILAMENT_START_END.value:
        controller.view.err_message( msg="Only the 'Rectangle (Filament Start-end)' visualization is enabled when picking filaments")
        return False

    if event.xdata is None or event.ydata is None or event.xdata < 0 or event.ydata < 0:
        controller.view.err_message( msg="Picked out of the image")
        return False

    if controller.view.plot and controller.view.plot.toolbar:
        # in case we are using the toolbar for zoom or pan
        return not controller.view.plot.toolbar.mode in ["zoom rect","pan/zoom"]

    return True

def onclick(controller,event:helper_GUI.plt_qtbackend)->None:
    """
    It is called when we click on the image
    :param controller: Boxmanager_controller obj
    :param event: the click event
    """

    controller.model.has_filament = controller.view.get_picking_filament_cb_text() == constants.Picking_cb.FILAMENT.value
    is_tomo = controller.model.type_case in [constants.Type_case.TOMO.value, constants.Type_case.TOMO_FOLDER.value]

    if not _check_onclick_validity(controller=controller, event=event):
        return

    # after starting picking is not possible to switch from picking box and filament. you have to reset
    controller.view.blur_picking_combobox(has_filament=controller.model.has_filament, is_importing=False)

    y = event.ydata - controller.model.boxsize / 2
    x = event.xdata - controller.model.boxsize / 2

    # load the sketches present in the current image (or slice)
    pure_filename = controller.get_current_filename(with_extension=False)
    controller.model.rectangles = controller.model.box_dictionary.get_sketches(f=pure_filename,
                                                                               n=controller.model.index_tomo,
                                                                               is_3D=controller.model.has_3d,  is_tracing = False)

    if controller.model.has_filament:

        if controller.view.control_modifier_is_clicked():
            remove_picked_filament(controller=controller, x=event.xdata, y=event.ydata,f=pure_filename, n=controller.model.index_tomo)
            # check or uncheck the related checkbox
            controller.view.check_item_checkbox(item=controller.view.tree.selectedItems()[0], check=len(controller.model.rectangles) > 0)
            return
        """

        # if we load from file and we want to add filament we have to insert the loaded filament in the picked_filament_list
        if self.picked_filament_list == list():
            if self.is_folder_3D_tomo and pure_filename in self.picked_filament_dictionary and self.index_3D_tomo in \
                    self.picked_filament_dictionary[pure_filename]:
                self.picked_filament_list = self.picked_filament_dictionary[pure_filename][self.index_3D_tomo]
            elif self.is_folder_3D_tomo is False and pure_filename in self.picked_filament_dictionary:
                self.picked_filament_list = self.picked_filament_dictionary[pure_filename]
            self.picked_filament_as_rect_dictionary = [fil.get_rect_sketch() for fil in self.picked_filament_list]
        """
        controller.model.picking_filament = True

        controller.model.current_pick_filament = helper_filament.picked_filament(box_size=controller.model.boxsize, is_tomo=is_tomo)
        controller.model.current_pick_filament.set_first(event=event)
        controller.view.check_item_checkbox(item=controller.view.tree.selectedItems()[0], check=True)
        controller.model.box_dictionary.add_picked_filament(still_picking=True,
                                                            fil = controller.model.current_pick_filament.get_rect_sketch(),
                                                            f = controller.get_current_filename(with_extension=False), n = controller.model.index_tomo)

        return

    box = helper.get_corresponding_box(x=x, y=y, get_low=False,
                                       rectangles=controller.model.rectangles,
                                       current_conf_thresh=controller.model.current_conf_thresh,
                                       box_size=controller.model.boxsize
                                       )

    if controller.view.control_modifier_is_clicked() or controller.view.meta_modifier_is_clicked():
        delete_box(controller=controller,box=box,f=pure_filename, n=controller.model.index_tomo)
    else:
        if controller.model.has_3d and box and box.only_3D_visualization:
            _, index = controller.model.box_dictionary.get_central_n(id_part=box.id,
                                                                       f=pure_filename, n=controller.model.index_tomo)
            controller.view.err_message(msg= f"for moving the box you have to click on its center, that is in slice number '{ index}'")
            return
        controller.model.moving_box = box
        if not controller.model.moving_box:
            # Delete lower confidence box if available
            box = helper.get_corresponding_box(x=x, y=y, get_low=True,
                                               rectangles=controller.model.rectangles,
                                               current_conf_thresh=controller.model.current_conf_thresh,
                                               box_size=controller.model.boxsize
                                               )
            if box:
                controller.model.rectangles.remove(box)
            # Create new box
            est_size=controller.model.boxsize

            rect = MySketch.MySketch(xy=(x, y), width=controller.model.boxsize, height=controller.model.boxsize,
                                     is_tomo=is_tomo, angle=0.0, est_size=est_size, confidence=1,
                                     only_3D_visualization=False, num_boxes=1, meta=None, z=None,
                                     linewidth=1, edgecolor="r", facecolor="none")

            # update dicts and controller.model.rectangles (it is a reference to controller.model.box_dictionary)
            controller.model.box_dictionary.counter+=1
            rect.id = controller.model.box_dictionary.counter
            controller.model.moving_box = rect
            controller.model.box_dictionary.add_sketch(sketch=rect, f=pure_filename, n=controller.model.index_tomo)

            # Add the patch to the Axes
            use_circle = controller.model.visualization == constants.Visualization_cb.CIRCLE.value
            controller.view.ax.add_patch(rect.getSketch(circle=use_circle))
            controller.view.ax.draw_artist(rect.getSketch(circle=use_circle))
            controller.view.fig.canvas.blit(controller.view.ax.bbox)

    # update_current=false because in case of deletion of 3D particle we could visualize the image where the center does not stay
    controller.view.update_tree_boxsizes(update_current=False, all_tomos=False)
    controller.view.update_global_counter(childid=None)

    controller.model.unsaved_changes = True

    # check or uncheck the related checkbox
    controller.view.check_item_checkbox(item=controller.view.tree.selectedItems()[0], check=len(controller.model.rectangles) > 0)

def is_picking_filament_enable(controller)->bool:
    """
    It is called in onmove and onrelease to detect is picking is enable
    :param controller: Boxmanager_controller obj
    :return False if picking is disabled
    """
    if controller.model.has_filament and controller.model.visualization != constants.Visualization_cb.RECT_FILAMENT_START_END.value:
        # we can pick only in rect-start visualization
        return False
    if controller.model.has_filament and controller.model.has_3d:
        # We cannot pick after importing TRACED filament
        return False
    if controller.view.plot and controller.view.plot.toolbar:
        # in case we are using the toolbar for zoom or pan
        return not controller.view.plot.toolbar.mode in ["zoom rect","pan/zoom"]
    return True

def onresize(controller, event:helper_GUI.plt_qtbackend)->None:
    """
    It is called when we resize the main window
    :param controller: Boxmanager_controller obj
    :param event: the click event
    """

    delete_all_patches(controller=controller)

    controller.view.fig.canvas.draw()
    controller.view.background_current = controller.view.fig.canvas.copy_from_bbox(controller.view.ax.bbox)
    controller.view.background_current_onmove = controller.view.background_current

    # draw the sketches present in the current image/slice
    if controller.model.rectangles:
        draw_all_patches(controller=controller)


def ondraw(controller, event:helper_GUI.plt_qtbackend)->None:
    """
    It is called when we use the toolbar
    :param controller: Boxmanager_controller obj
    :param event: the click event
    """
    if controller.view.zoom_update:
        controller.view.zoom_update = False
        controller.view.background_current = controller.view.fig.canvas.copy_from_bbox(controller.view.ax.bbox)
        controller.view.background_current_onmove = controller.view.fig.canvas.copy_from_bbox(controller.view.ax.bbox)
        # draw the sketches again
        draw_all_patches(controller=controller)
        controller.view.fig.canvas.blit(controller.view.ax.bbox)

def onmove(controller, event:helper_GUI.plt_qtbackend)->None:
    """
    It is called when we move the cursor over the image
    :param controller: Boxmanager_controller obj
    :param event: the click event
    """

    if event.inaxes != controller.view.ax or controller.view.control_modifier_is_clicked():
        return

    if not is_picking_filament_enable(controller=controller):
        return

    if event.button == 1:
        if controller.view.get_picking_filament_cb_text() == constants.Picking_cb.PARTICLE.value and controller.model.moving_box:
            # in tomo case i remove the 3D visualization BUT not the 2D original sketch that is represented by  'controller.model.moving_box'
            if controller.model.has_3d and not controller.model.is_moving:
                #remove the whole 3D particle
                controller.model.box_dictionary.delete_3D(id_part=controller.model.moving_box.id,
                                                          f=controller.get_current_filename(False),
                                                          n=controller.model.index_tomo)
                # re insert ONLY its center. the 3D ll be create in the 'onrelease'
                controller.model.box_dictionary.particle3D_dictionary[controller.get_current_filename(with_extension=False)][controller.model.index_tomo].append(controller.model.moving_box)
                controller.model.is_moving = True

            is_circle = controller.model.visualization == constants.Visualization_cb.CIRCLE.value

            rect_width = controller.model.moving_box.getSketch().get_width()
            x = event.xdata - rect_width / 2
            y = event.ydata - rect_width / 2

            # set the new x,y coordinate for the rectangle sketch
            controller.model.moving_box.getSketch(circle=False).set_x(x)
            controller.model.moving_box.getSketch(circle=False).set_y(y)
            controller.model.moving_box.set_xy((x, y), circle=False)

            # set the new x,y coordinate for the circle sketch
            controller.model.moving_box.getSketch(circle=True).center = (event.xdata, event.ydata)
            controller.model.moving_box.set_xy((event.xdata, event.ydata), circle=True)

            # restore the old background image (without sketches) set in the 'event_image_changed'
            if controller.view.background_current:
                controller.view.fig.canvas.restore_region(controller.view.background_current)

            # Add al the current patches to the Axes. NB: the moving_box is a reference to a box present into this var
            for b in controller.model.rectangles:
                controller.view.ax.add_patch(b.getSketch(circle=is_circle ))
                controller.view.ax.draw_artist(b.getSketch(circle=is_circle ))

            controller.view.fig.canvas.blit(controller.view.ax.bbox)
            return
        elif controller.view.get_picking_filament_cb_text() == constants.Picking_cb.FILAMENT.value and controller.model.picking_filament:
            controller.model.current_pick_filament.end_fil = [event.xdata,event.ydata]

            # clean the image removing the patches and save as current background
            delete_all_patches(controller=controller)
            if controller.view.background_current_onmove:
                controller.view.fig.canvas.restore_region(controller.view.background_current_onmove)

            # remove current filament and its center line
            for i in range(len(controller.model.rectangles) - 1, -1, -1):
                if controller.model.rectangles and controller.model.rectangles[i].id in [controller.model.box_dictionary.counter, constants.CENTER_LINE_ID]:
                    del controller.model.rectangles[i]

            # create center_line
            center_line = helper_filament.picked_filament(box_size=1)
            center_line.end_fil = controller.model.current_pick_filament.end_fil
            center_line.begin_fil = controller.model.current_pick_filament.begin_fil
            controller.model.rectangles.append(center_line.get_rect_sketch(sketch_id=constants.CENTER_LINE_ID))
            # fill again the current image with boxes
            controller.model.box_dictionary.add_picked_filament(still_picking=True,
                fil=controller.model.current_pick_filament.get_rect_sketch(),
                f=controller.get_current_filename(with_extension=False), n=controller.model.index_tomo)
            # draw the sketches again
            draw_all_patches(controller=controller)
            controller.view.fig.canvas.blit(controller.view.ax.bbox)


def onrelease(controller,event:helper_GUI.plt_qtbackend)->None:
    """
    It is called when we release the mouse button
    :param controller: Boxmanager_controller obj
    :param event: the click event
    """
    if not is_picking_filament_enable(controller=controller):
        return

    # clean the image removing the patches and save as current background
    delete_all_patches(controller=controller)
    if controller.view.background_current_onmove:
        controller.view.fig.canvas.restore_region(controller.view.background_current_onmove)


    pure_filename=controller.get_current_filename(with_extension=False)
    if controller.view.get_picking_filament_cb_text() == constants.Picking_cb.PARTICLE.value:
        # in tomo case i have to create the 'moving_box' 3D structure
        if controller.model.has_3d and controller.model.is_moving:
            controller.model.is_moving = False
            controller.model.box_dictionary.create_3D(sketch=controller.model.moving_box,
                                                      f=pure_filename,
                                                      n=controller.model.index_tomo,is_start_end_filament=False)

        controller.model.moving_box = None
    elif not controller.model.removing_picked_filament:
        # remove current center line
        for i in range(len(controller.model.rectangles) - 1, -1, -1):
            if controller.model.rectangles and controller.model.rectangles[i].id == constants.CENTER_LINE_ID:
                del controller.model.rectangles[i]
                break

        controller.model.box_dictionary.add_sketches(sketches=controller.model.current_pick_filament.get_sketches(box_distance=controller.model.box_distance_filament, as_bbox=False, z=controller.model.index_tomo),
                                                     f=pure_filename, n=controller.model.index_tomo, id_fil=controller.model.box_dictionary.counter)
        #update the total counter
        controller.model.box_dictionary.update_counter(n=1)
        controller.model.picking_filament = False

    # draw the sketches again
    draw_all_patches(controller=controller)
    controller.view.fig.canvas.blit(controller.view.ax.bbox)
    # update the counters
    controller.view.update_tree_boxsizes()
    controller.view.update_global_counter()
    controller.model.removing_picked_filament = False

def myKeyPressEvent(controller:boxmanager_controller, event:helper_GUI.plt_qtbackend)->None:
    """
    It is called when we press the keyboard
    :param controller: Boxmanager_controller obj
    :param event: the press keyboard event
    """
    def get_number_current_tomo():
        for index,name in enumerate(controller.model.tot_frames.keys()):
            if name == current_tomo_name:
                return index
        return 0

    if event.name == "key_press_event":
        is_tomo = controller.model.type_case in [constants.Type_case.TOMO.value, constants.Type_case.TOMO_FOLDER.value]
        if event.key == "h":
            h_keypress(controller=controller)
        elif is_tomo and event.key in ["down","up"]:
            current_tomo_name = controller.get_current_filename(with_extension=False)
            controller.model.index_tomo = max(0,controller.model.index_tomo-1) if event.key == "down" else min(controller.model.tot_frames[current_tomo_name]-1,controller.model.index_tomo+1)
            controller.view.setCurrentItem_viewTree(is_tomo_folder=controller.model.type_case == constants.Type_case.TOMO_FOLDER.value, number_tomo=get_number_current_tomo(), z=controller.model.index_tomo)
            event_image_changed(controller=controller,root_tree_item=controller.view.tree.currentItem())


def h_keypress(controller:boxmanager_controller)->None:
    currentfile = controller.get_current_filename(with_extension=False)

    # delete the sktches
    if controller.model.rectangles:
        delete_all_patches(controller=controller)
        controller.view.fig.canvas.draw()
        controller.view.background_current = controller.view.fig.canvas.copy_from_bbox(controller.view.ax.bbox)
    # switch the toggle value
    controller.model.toggle = not controller.model.toggle

    # I have to colelct the data to show
    controller.model.rectangles = list() if controller.model.toggle else controller.model.box_dictionary.get_sketches(
        f=currentfile, n=controller.model.index_tomo, is_3D=controller.model.has_3d,  is_tracing = False)

    # plot the sketches
    if controller.model.rectangles:
        draw_all_patches(controller=controller)
        controller.view.fig.canvas.blit(controller.view.ax.bbox)

def use_estimated_size_changed(controller:boxmanager_controller)->None:
    """
    It changes the value of the "box size" filling the 'box size' line in the visualization tab.
    It is usable only after loading data from cbox file. It gets the new "box size" from these files
    Pay attention: each particle will have its own estimated size
    It is triggered pressing enter (when the cursor is into the line) or button 'set'
    """
    controller.model.use_estimated_size = controller.view.use_estimated_size_checkbox.isChecked()

    # since this optionis basically a resize_box we do not need to run 'h_keypress' or unset toggle flag

    # when we uncheck we resize with the last box_size value
    if not controller.model.use_estimated_size:
        box_size_changed(controller=controller,box_size_from_file=None)

    #remove the old patches, they are already saved in controller.model.rectangles
    if controller.model.rectangles:
        delete_all_patches(controller=controller)
        if controller.view.background_current:
            controller.view.fig.canvas.restore_region(controller.view.background_current)

    # resize the whole 3D dict
    if controller.model.has_3d:
        controller.model.box_dictionary.resize_3D(new_boxsize=controller.model.boxsize,  old_boxsize=None, use_estimated_size=True)

    if controller.model.rectangles:
        # collect the data to plot
        if controller.model.has_3d:
            controller.model.rectangles = controller.model.box_dictionary.get_sketches(
                f=controller.get_current_filename(with_extension=False),
                n=controller.model.index_tomo, is_3D=True,  is_tracing = False)
        else:
            controller.model.rectangles = controller.model.box_dictionary.resized_and_get_sketches(
            new_boxsize=controller.model.boxsize, f=controller.get_current_filename(with_extension=False),
                n=controller.model.index_tomo,use_estimated_size=controller.model.use_estimated_size)

        # draw the sketches present in the current image/slice
        draw_all_patches(controller=controller)
        controller.view.fig.canvas.blit(controller.view.ax.bbox)

def box_size_changed(controller:boxmanager_controller, box_size_from_file:int = None)->None:
    """
    It changes the value of the "box size" filling the 'box size' line in the visualization tab.
    It is triggered pressing enter (when the cursor is into the line) or button 'set'
    :param controller: Boxmanager_controller obj
    :param box_size_from_file: Not None when we load from file the boxes and we want to set the value
    """
    # set the boxsize in case we loaded from file (NB: it has nothing to do with estimate size, it is box.w)
    if controller.view.is_updating_params:
        controller.view.is_updating_params=False
        controller.model.boxsize = box_size_from_file
        controller.view.boxsize_line.setText(str(box_size_from_file))
        controller.view.line_setText(line=controller.view.boxsize_line, value=str(box_size_from_file))
        return

    # After clicking 'h' the boxes disappear, before acting i have to untoggle
    if controller.model.toggle:
        h_keypress(controller=controller)

    # if the use_estimated_size checkbox is checked we can change the boxsize as always but we have to uncheck the checkbox
    if controller.model.use_estimated_size:
        controller.view.check_use_estimated_size_checkbox(check=False) # call to 'use_estimated_size_changed'


    old_value = controller.model.boxsize
    try:
        value = int(float(controller.view.boxsize_line.text()))
        if value <= 0:
            controller.view.line_setText(line=controller.view.boxsize_line, value=str(old_value))
            controller.view.err_message(msg="Invalid value. It has to be a positive number")
            return
        controller.model.boxsize = value
    except ValueError:
        controller.view.line_setText(line=controller.view.boxsize_line, value=str(old_value))
        controller.view.err_message( msg="Invalid value. It has to be a positive number")
        return

    current_filename = controller.get_current_filename(with_extension=False)

    # update the params var. Since this is a high consuming time operation we perform it only in the current tomo
    if controller.model.type_case == constants.Type_case.TOMO_FOLDER.value:
        controller.model.params[current_filename].box_size = controller.model.boxsize

    if controller.model.rectangles:
        #remove the old patches, they are already saved in controller.model.rectangles
        delete_all_patches(controller=controller)
        if controller.view.background_current:
            controller.view.fig.canvas.restore_region(controller.view.background_current)

    # resize the whole 3D dict
    if controller.model.has_3d:
        controller.model.box_dictionary.resize_3D(new_boxsize=controller.model.boxsize,old_boxsize=old_value,use_estimated_size=False)

    if controller.model.rectangles:
        # collect the data to plot
        if controller.model.has_3d:
            controller.model.rectangles = controller.model.box_dictionary.get_sketches(f=controller.get_current_filename(with_extension=False),
                                                                            n=controller.model.index_tomo,is_3D = True,  is_tracing = False)

        elif not controller.model.has_3d or controller.model.is_cbox_untraced:
            controller.model.rectangles = controller.model.box_dictionary.resized_and_get_sketches(new_boxsize=controller.model.boxsize,
                                                                                               f=controller.get_current_filename(with_extension=False),
                                                                                                   n=controller.model.index_tomo,
                                                                                                use_estimated_size=controller.model.use_estimated_size)

        # draw the sketches present in the current image/slice
        draw_all_patches(controller=controller)
        controller.view.fig.canvas.blit(controller.view.ax.bbox)


def box_distance_filament_changed(controller:boxmanager_controller, box_distance:int = 1)->None:
    # After clicking 'h' the boxes disappear, before acting i have to untoggle
    if controller.model.toggle:
        h_keypress(controller=controller)

    old_value=controller.model.box_distance_filament
    try:
        value = int(float(controller.view.box_distance_filament_line.text()))
        if value <= 0:
            controller.view.line_setText(line=controller.view.box_distance_filament_line, value=str(old_value))
            controller.view.err_message(msg="Invalid value. It has to be a positive number")
            return
        controller.model.box_distance_filament = value
    except ValueError:
        controller.view.line_setText(line=controller.view.box_distance_filament_line, value=str(old_value))
        controller.view.err_message( msg="Invalid value. It has to be a positive number")
        return

    if controller.model.rectangles:
        #remove the old patches, they are already saved in controller.model.rectangles
        delete_all_patches(controller=controller)
        if controller.view.background_current:
            controller.view.fig.canvas.restore_region(controller.view.background_current)

    current_filename =controller.get_current_filename(with_extension = False)
    controller.model.box_dictionary.update_box_distance_dictionary(new_box_distance=controller.model.box_distance_filament,f=current_filename,n=controller.model.index_tomo)

    controller.model.rectangles = controller.model.box_dictionary.change_box_distance(new_box_distance=controller.model.box_distance_filament,f=current_filename,n=controller.model.index_tomo, is_tracing=False)

    # draw the sketches present in the current image/slice
    draw_all_patches(controller=controller)
    controller.view.fig.canvas.blit(controller.view.ax.bbox)

def visualization_changed(controller:boxmanager_controller)->None:
    """
    It changes the value of the "visualization_combobox" changing the 'visualization' combobox in the visualization tab.
    It is triggered when the combobox is changed
    """
    # After clicking 'h' the boxes disappear, before acting i have to untoggle
    if controller.model.toggle:
        h_keypress(controller=controller)

    if controller.model.rectangles:
        #remove the old patches (delete_all_patches recognizes the correct visualization), they are already saved in controller.model.rectangles
        delete_all_patches(controller=controller)
        # restore the canvas
        if controller.view.background_current:
            controller.view.fig.canvas.restore_region(controller.view.background_current)

    old_vis = controller.model.visualization
    current_filename = controller.get_current_filename(with_extension=False)
    controller.model.visualization = controller.view.get_visualization_cb_text()
    controller.model.box_dictionary.update_visualization_dictionary(new_visualization=controller.model.visualization,
                                                                    f=current_filename, n=controller.model.index_tomo)

    # If filaments mode i have to set the rectangles
    if controller.model.has_filament and constants.Visualization_cb.RECT_FILAMENT_START_END.value in [old_vis,controller.model.visualization]:
        controller.model.box_dictionary.fil_start_end_vis = controller.model.visualization == constants.Visualization_cb.RECT_FILAMENT_START_END.value
        controller.model.rectangles = controller.model.box_dictionary.get_sketches(f=current_filename,
                                                        n=controller.model.index_tomo, is_3D =controller.model.has_3d,  is_tracing = False)

    # draw again the patches, (draw_all_patches recognizes the correct visualization)
    draw_all_patches(controller=controller)
    controller.view.fig.canvas.blit(controller.view.ax.bbox)

# FILTERING TAB OPTION
def apply_filter(controller:boxmanager_controller)->None:
    """
    It applies the lowpass filter present in filtering tab.
    :param controller: Boxmanager_controller obj
    """
    old_value=controller.model.filter_freq
    try:
        controller.model.filter_freq = float(controller.view.filter_line.text())
    except ValueError:
        controller.view.err_message(msg="Invalid character. The Frequency value has to be a number between 0 and 0.5")
        return

    if 0 <=controller.model.filter_freq < 0.5:
        import cryolo.lowpass as lp
        if controller.model.type_case == constants.Type_case.SPA.value:
            img = lp.filter_single_image(img_path=controller.model.current_image_path,
                                         filter_cutoff=controller.model.filter_freq, resize_to_shape=None)
            im_type = constants.Filetype_name[path.splitext(controller.model.current_image_path)[1].replace(".","").upper()].value
            img = helper.normalize_and_flip(img=img, file_type=im_type)
        else:
            img = helper.read_image(file_path=controller.model.current_image_path, use_mmap=True)
            img = lp.filter_single_image_from_np_array(image=img[controller.model.index_tomo, :, :],
                                                       filter_cutoff=controller.model.filter_freq, resize_to_shape=None,
                                                       channel_specifiction=0)
        img = np.squeeze(img)
        delete_all_patches(controller=controller)
        controller.view.im.set_data(img)
        controller.view.fig.canvas.draw()
        controller.view.background_current = controller.view.fig.canvas.copy_from_bbox(controller.view.ax.bbox)
        controller.view.background_current_onmove = controller.view.fig.canvas.copy_from_bbox(controller.view.ax.bbox)
        draw_all_patches(controller=controller)
    else:
        controller.model.filter_freq = old_value
        controller.view.err_message( msg="Frequency has to be between 0 and 0.5")


# DELETE AND DRAW ON IMAGE FUNCTIONS
def delete_all_patches(controller:boxmanager_controller)->None:
    """
    Deletes all the patches present in current axes
    Since model.rectangle is a reference to a list of particle_dictionary dict it has to fill in the appropriate way
    :param controller: Boxmanager_controller obj
    """

    controller.view.ax.patches=list()

def draw_all_patches(controller:boxmanager_controller)->None:
    """
    Draws all the patches present in model.rectangle var
    Since model.rectangle is a reference to a list of particle_dictionary dict it has to fill in the appropriate way
    :param controller: Boxmanager_controller obj
    """
    current_conf_thresh, upper_size_thresh, lower_size_thresh, current_num_boxes_thresh = controller.model.get_threshold_params(filename=None)
    has_filament = controller.model.picking_combobox_value == constants.Picking_cb.FILAMENT.value
    is_circle = controller.model.visualization in [constants.Visualization_cb.CIRCLE.value,constants.Visualization_cb.CIRCLE_SEGMENTED.value]
    visible_sketches = [box.getSketch(circle=is_circle) for box in controller.model.rectangles if
                     helper.check_if_should_be_visible(box=box, current_conf_thresh=current_conf_thresh, upper_size_thresh=upper_size_thresh, lower_size_thresh=lower_size_thresh,
                                                       num_boxes_thresh=current_num_boxes_thresh, is_filament=has_filament)]

    for sketch in visible_sketches:
        if not sketch.get_visible():
            sketch.set_visible(True)

        controller.view.ax.add_patch(sketch)
        controller.view.ax.draw_artist(sketch)

def delete_box(controller:boxmanager_controller, box:MySketch.MySketch, f:str, n:int)->None:
    """
    Removed the selected box. It is identified via GUI clicking 'ctrl+left_button_mouse' into the box
    Since model.rectangle is a reference to a list of particle_dictionary dict it has to fill in the appropriate way
    :param controller: Boxmanager_controller obj
    :param box: sketch to remove
    :param f: filename
    :param n: number of slice
    """
    if box:
        # delete the current sketches from the figure
        delete_all_patches(controller=controller)

        # restore the canvas
        if controller.view.background_current:
            controller.view.fig.canvas.restore_region(controller.view.background_current)

        # removed the selected sketch from the model.box_dictionary dicts and 'controller.model.rectangles' list
        # in case we do not clicked on the central 2D section of 3D particle we have to calculate the boxes list again
        # in case we do not clicked on the central 2D section of 3D particle we have to calculate the central box
        if controller.model.has_3d and box.only_3D_visualization:
            box, index = controller.model.box_dictionary.get_central_n(id_part=box.id,f=f, n=controller.model.index_tomo)
            boxes = controller.model.box_dictionary.get_sketches(f=f, n=index, is_3D=False,  is_tracing = False)
            controller.model.box_dictionary.delete_sketch_by_index(index=boxes.index(box), f=f, n=index)
            # uncheck the checkbox if there are no more boxes there
            if not boxes:
                controller.view.set_checkstate_tree_leafs(item=controller.view.tree.invisibleRootItem(), entries=[str(index)],
                                      state=controller.view.checkbox_state_value(checked=False))
        else:
            controller.model.box_dictionary.delete_sketch_by_index(index=controller.model.rectangles.index(box),f=f,n=n)

        # draw the sketches again
        draw_all_patches(controller=controller)
        controller.view.fig.canvas.blit(controller.view.ax.bbox)

def is_click_into_box(fil:MySketch.MySketch,x:float,y:float,boxsize:int,is_tomo:bool)->bool:
    """
    Return True if thce click (x,y) is into the filament
    :param fil: filamnent MySketch obj
    :param x: x coordinate of the mouse click
    :param y: y coordinate of the mouse click
    :param boxsize: boxsize
    :param is_tomo: True if it is a tomo case
    """
    filament = helper_filament.picked_filament(box_size=boxsize,is_tomo=is_tomo)
    filament.create_from_rect(rect=fil)
    rotated_corners=filament.get_rotated_coordinate_rect()
    rotated_point = filament.get_R() @ np.array((x, y)).T
    max_value = np.amax(rotated_corners,axis=1)
    min_value = np.amin(rotated_corners, axis=1)
    return min_value[0]< rotated_point[0] <max_value[0] and min_value[1]<rotated_point[1]<max_value[1]

def remove_picked_filament(controller:boxmanager_controller, x:float,y:float, f:str, n:int)->None:
    """
    Removed the selected filament. It is identified via GUI clicking 'ctrl+left_button_mouse' into the box
    Since model.rectangle is a reference to a list of particle_dictionary dict it has to fill in the appropriate way
    :param controller: Boxmanager_controller obj
    :param x: coordinate x of the click
    :param y: coordinate y of the click
    :param f: filename
    :param n: number of slice
    """
    controller.model.removing_picked_filament = True
    is_tomo = controller.model.type_case in [constants.Type_case.TOMO.value, constants.Type_case.TOMO_FOLDER.value]
    for index, fil in enumerate(controller.model.rectangles):
        if is_click_into_box(fil,x,y,boxsize=controller.model.boxsize,is_tomo=is_tomo):
            if controller.model.has_filament:
                controller.model.box_dictionary.delete_filemant_by_index(index=index, f=f,n=n)
            else:
                controller.model.box_dictionary.delete_sketch_by_index(index=index, f=f,n=n)
            controller.model.unsaved_changes = True
            # restore the canvas
            if controller.view.background_current:
                controller.view.fig.canvas.restore_region(controller.view.background_current)

            # draw the sketches again
            draw_all_patches(controller=controller)
            controller.view.fig.canvas.blit(controller.view.ax.bbox)

            # update the counters
            controller.view.update_tree_boxsizes(update_current=False)
            controller.view.update_global_counter(childid=None)
            return




"""
In this this file are defined a list of various helper functions (and an helper class):
-) create parser: manage the parser
"""

import argparse
import numpy as np
from math import isnan
from os import listdir,path
from mrcfile import mmap as mrcfile_mmap        #it is not in the setup.py because it is already installed by cryolo

from cryoloBM import MySketch, constants, helper_GUI
from cryolo import  utils,imagereader

from cryoloBM import boxmanager_controller # for type hinting purpose
from typing import Union,List,Dict,Tuple

def create_parser():
    """
    Create the parser
    """
    parser = argparse.ArgumentParser(
        description="Train and validate crYOLO on any dataset",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("-i", "--image_dir", help="Path to image directory.")
    parser.add_argument("-b", "--box_dir", help="Path to box directory.")
    parser.add_argument("--wildcard",
                           help="Wildcard for selecting spcific images (e.g *_new_*.mrc)")
    parser.add_argument("-t", "--is_tomo", action="store_true",
                           help="Flag for specifying that the directory or the path  contains tomogram(s).")
    return parser

def set_current_image_path(controller:boxmanager_controller, current_image_path:str)->None:
    """
    Set the current image path for the model and view
    :param controller: Boxmanager_model obj
    :param current_image_path: the current image path
    """
    controller.model.current_image_path = current_image_path
    controller.view.current_image_path = current_image_path

def is_cbox_untraced(b:utils.BoundBox)->bool:
    """
    :param b: loaded box of a cbox file
    return False True if the file is cobx_untraced
    """

    # micrograph case has always b.depth == None and crYOLO will never spawn a untraced file for micrograph case
    if b.depth is not None and isnan(b.depth) is False:
        return int(b.depth) == 1
    return False


def get_corresponding_box(x:int, y:int, rectangles:List[MySketch.MySketch], current_conf_thresh:int, box_size:int, get_low:bool=False)->Union[MySketch.MySketch,None]:
    a = np.array([x, y])

    for box in rectangles:
        b = np.array(box.get_xy())
        dist = np.linalg.norm(a - b)
        if get_low:
            if dist < box_size / 2 and box.get_confidence() < current_conf_thresh:
                return box
        else:
            if dist < box_size / 2 and box.get_confidence() > current_conf_thresh:
                return box
    return None


def read_image(file_path:str, use_mmap:bool=False)->np.ndarray:
    im_type = constants.Filetype_name[path.splitext(file_path)[1].replace(".","").upper()].value

    img = imagereader.image_read(file_path, use_mmap=use_mmap)
    img = normalize_and_flip(img, im_type)
    return img


def normalize_and_flip(img:np.ndarray, file_type:int)->np.ndarray:
    if file_type == 0:
        # JPG
        img = np.flip(img, 0)
    if file_type == 1 or file_type == 2:
        # tif /mrc
        if not np.issubdtype(img.dtype, np.float32):
            img = img.astype(np.float32)
        if len(img.shape) == 3:
            img = np.flip(img, 1)
        else:
            img = np.flip(img, 0)
        mean = np.mean(img)
        sd = np.std(img)
        img = (img - mean) / sd
        #img[img > 3] = 3
        #img[img < -3] = -3
    return img

def convert_list_bbox_to_sketch(boxes_list:List[utils.BoundBox], out_dict:bool=True, col:str ='r',fil_id=None)->Union[Dict[int,MySketch.MySketch],List[MySketch.MySketch]]:
    """
    Convert the list of Bbox (see 'cryolo.grouping3d.do_tracing' for the syntax), in a dict of sketches
    :param boxes_list: list of BoundingBox instances
    :param out_dict: If True convert in dict otherwise in list
    :param col: color of the edge
    :param fil_id: id filament, used for saving on file purpose and some conversion visualization
    :return dict of sketches (k = frame number, v = list of sketches)
    """
    out = dict() if out_dict else list()

    num_boxes = 1
    for b in boxes_list:
        if 'num_boxes' in b.meta:
            num_boxes = b.meta["num_boxes"]

        sketch = MySketch.MySketch(xy=(b.x,b.y), width=b.w, height=b.h,
                                   is_tomo=True, angle=0.0, est_size=b.w,
                                   confidence=b.c, only_3D_visualization=False,
                                   num_boxes=num_boxes, meta=None, z=b.z, id=fil_id,
                                   linewidth=1, edgecolor=col, facecolor="none")
        if out_dict is False:
            out.append(sketch)
        elif b.z in out:
            out[int(b.z)].append(sketch)
        else:
            out[int(b.z)] = [sketch]
    return out


class Params:
    """
    Used only in tomo folder case
    We storage the 'thresholding' and 'tracing' params in order to keep the state for each tomo
    """
    def __init__(self, tomo_name:str, filter_freq:float, box_size:int, upper_size_thresh:float, lower_size_thresh:float, conf_thresh:float, num_boxes_thresh:int,
                 min_edge_weight:int, search_range:int, memory:int, min_length:int, win_size:int)->None:
        self.tomo_name = tomo_name
        self.filter_freq = filter_freq
        self.box_size = box_size
        self.upper_size_thresh = upper_size_thresh
        self.lower_size_thresh = lower_size_thresh
        self.conf_thresh = conf_thresh
        self.num_boxes_thresh = num_boxes_thresh
        self.min_edge_weight = min_edge_weight
        self.search_range = search_range
        self.memory=  memory
        self.min_length = min_length
        self.win_size =win_size

    def reset_to_default_values(self)->None:
        """
        Reset to default value. Win_size default value is the box_size
        """
        self.filter_freq = constants.Default_settings_value.DEFAULT_FILTER_FREQ.value
        self.upper_size_thresh = constants.Default_settings_value.DEFAULT_UPPER_SIZE_THRESH.value
        self.lower_size_thresh = constants.Default_settings_value.DEFAULT_LOWER_SIZE_THRESH.value
        self.conf_thresh = constants.Default_settings_value.DEFAULT_CURRENT_CONF_THRESH.value
        self.num_boxes_thresh = constants.Default_settings_value.DEFAULT_CURRENT_NUM_BOXES_THRESH.value
        self.min_edge_weight = constants.Default_settings_value.DEFAULT_MIN_EDGE_WEIGHT.value
        self.search_range = constants.Default_settings_value.DEFAULT_SEARCH_RANGE.value
        self.memory = constants.Default_settings_value.DEFAULT_MEMORY.value
        self.min_length = constants.Default_settings_value.DEFAULT_MIN_LENGTH.value
        self.win_size = constants.Default_settings_value.DEFAULT_BOX_SIZE.value

    def has_same_tracing_params(self,param, has_filament:bool = False)->bool:
        """
        :param param: a Params obj
        :param has_filament: True if we analyze filament
        Return True if param has the same tracing parameters
        """
        min_cond = self.min_length == param.min_length and self.memory == param.memory and self.search_range  == param.search_range
        if has_filament:
            return min_cond and self.min_edge_weight == param.min_edge_weight and self.win_size == param.win_size
        return self.min_length == param.min_length and self.memory == param.memory and self.search_range  == param.search_range


def get_only_files(dir_path:str,wildcard:str,is_list_tomo:bool)->List[str]:
    """
    generate list of files in 'dir_path' path
    :param dir_path: path to the folder
    :param wildcard: using wildcard
    :param is_list_tomo: True if folder of 3D tomo
    :return: list of valid files in 'dir_path'
    """
    onlyfiles = [
        f
        for f in sorted(listdir(dir_path))
        if path.isfile(path.join(dir_path, f))
    ]

    if wildcard:
        import fnmatch
        onlyfiles = [
            f
            for f in sorted(listdir(dir_path))
            if fnmatch.fnmatch(f, wildcard)
        ]

    if is_list_tomo is False:
        onlyfiles = [
            i
            for i in onlyfiles
            if not i.startswith(".")
               and i.endswith((".jpg", ".jpeg", ".png", ".mrc", ".mrcs", ".tif", ".tiff", ".rec"))]
    else:
        onlyfiles_all = [i for i in onlyfiles if not i.startswith(".") and i.endswith((".mrc", ".mrcs",".rec"))]
        onlyfiles.clear()
        for f in onlyfiles_all:
            with mrcfile_mmap(path.join(dir_path, f), permissive=True, mode="r") as mrc:
                if len(mrc.data.shape) == 3:
                    onlyfiles.append(f)
                mrc.close()

    return onlyfiles

def list_files_in_folder(controller:boxmanager_controller,is_list_tomo:bool=False)->Union[Tuple[helper_GUI.QtG.QTreeWidgetItem, List[str], List[helper_GUI.QtG.QTreeWidgetItem]],Tuple[bool,bool,bool]]:
    """
    :param controller: Boxmanager_controller obj
    :param is_list_tomo: True if folder of 3D tomo
    :return root: the QTreeWidgetItem
    :return onlyfiles: list of valid files
    :return all_items: items of the root (one for each valid file)
    """
    if controller.model.image_folder_path != "" and controller.model.image_folder_path is not None:
        title = path.join(str(controller.model.image_folder_path), controller.model.wildcard) if controller.model.wildcard else str(controller.model.image_folder_path)
        root = controller.view.create_tree_widget_item( title=title)
        controller.view.check_item_checkbox(item=root, check=False)

        controller.view.reset_tree(root, title)
        controller.model.set_rectangles(rectangles=[])

        onlyfiles = get_only_files(controller.model.image_folder_path,controller.model.wildcard,is_list_tomo)
        all_items = [controller.view.create_tree_widget_item( title=file) for file in onlyfiles]

        if len(all_items) > 0:
            pd = controller.view.create_progress_dialog(title="Load images")
            pd.show()
            for item_index, item in enumerate(all_items):
                #controller.view.run_in_same_thread()       # sometimes it crashes
                pd.setValue(int((item_index+1)*100/len(all_items)))
                controller.view.check_item_checkbox(item=item, check=False)
                root.addChild(item)
            pd.close()
        return root,onlyfiles,all_items
    return False,False,False


def get_all_loaded_filesnames(root:helper_GUI.QtG.QTreeWidgetItem)->List[str]:
    """
    get the list of the loaded file
    :param root: QtG.QTreeWidget obj
    :return: list of filenames
    """
    return [path.splitext(root.child(i).text(0))[0] for i in range(root.childCount())]

def check_if_should_be_visible( box:MySketch.MySketch, current_conf_thresh:float, upper_size_thresh:float, lower_size_thresh:float, num_boxes_thresh:int = 0, is_filament:bool = True)->bool:
    # Cryolo returns estimated value = NaN for the filament mode. hence in these case we filter only in according with its num_boxes and confidence thresholding
    est_size = upper_size_thresh if is_filament else box.get_est_size()
    if isnan(est_size) :
        return True
    cond = [box.get_confidence() > current_conf_thresh or isnan(box.get_confidence()),
            upper_size_thresh >= est_size >= lower_size_thresh ,
            box.num_boxes > num_boxes_thresh]
    return all(cond)

def create_sketch(sketch:MySketch.MySketch, delta:int, size:int, is_start_end_filament:bool)->MySketch.MySketch:
    """
    Create a new 'smaller' sketch of the given 'sketch' for creating the 3D visualization
    """
    est_size = sketch.get_est_size(False)
    if not est_size:
        est_size = sketch.get_width()
    xy = sketch.get_xy(circle = False)
    width = sketch.get_width() if is_start_end_filament else size
    return MySketch.MySketch(xy=(xy[0] + delta, xy[1] + delta), width=width, height=size, is_tomo=sketch.get_is_tomo(),
                             angle=sketch.get_angle(), est_size=est_size, confidence=sketch.get_confidence(),
                             only_3D_visualization=True, num_boxes=sketch.num_boxes, meta=None, z=None, id=sketch.id,
                             linewidth=1, edgecolor=sketch.getSketch().get_edgecolor(), facecolor="none")

def load_tomo_params(controller:boxmanager_controller,pure_filename:str)->None:
    """
    reset the param f the given tomo, it is basically used in '_event_image_changed'
    :param controller
    :param pure_filename: name of the tomo
    """
    #self.is_loading_max_min_sizes = True
    controller.view.is_updating_params = True
    controller.model.upper_size_thresh = controller.model.params[pure_filename].upper_size_thresh
    controller.view.line_setText(line=controller.view.upper_size_thresh_line, value=str(controller.model.upper_size_thresh))
    controller.model.lower_size_thresh = controller.model.params[pure_filename].lower_size_thresh
    controller.view.line_setText(line=controller.view.lower_size_thresh_line,value=str(controller.model.lower_size_thresh))
    controller.model.current_conf_thresh = controller.model.params[pure_filename].conf_thresh
    controller.view.line_setText(line=controller.view.conf_thresh_line,value=str(controller.model.current_conf_thresh))
    controller.model.current_num_boxes_thresh =controller.model.params[pure_filename].num_boxes_thresh
    controller.view.line_setText(line=controller.view.num_boxes_thresh_line,value=str(controller.model.current_num_boxes_thresh))
    if controller.model.is_cbox_untraced:
        controller.model.memory = controller.model.params[pure_filename].memory
        controller.view.line_setText(line=controller.view.memory_line,value=str(controller.model.memory))
        controller.model.search_range = controller.model.params[pure_filename].search_range
        controller.view.line_setText(line=controller.view.search_range_line,value=str(controller.model.search_range))
        controller.model.min_length = controller.model.params[pure_filename].min_length
        controller.view.line_setText(line=controller.view.min_length_line, value=str(controller.model.min_length))
        if controller.model.has_filament:
            controller.model.win_size = controller.model.params[pure_filename].win_size
            controller.view.line_setText(line=controller.view.win_size_line,value=str(controller.model.win_size))
            controller.model.min_edge_weight = controller.model.params[pure_filename].min_edge_weight
            controller.view.line_setText(line=controller.view.min_edge_weight_line,value=str(controller.model.min_edge_weight))
    controller.view.is_updating_params = False


"""
In this tool we use the Model View Controller (MVC) design pattern.
The MCV can be considered an approach to distinguish between the data model, processing control and the user interface.
It neatly separates the graphical interface displayed to the user from the code that manages the user actions.

In this file we define the 'Boxmanager_model' which deals with the data
"""

from cryoloBM import constants,box_dictionary,helper
from cryoloBM import MySketch   # for type hinting purpose
from typing import List

class Boxmanager_model:
    def __init__(self, image_dir:str, box_dir:str, wildcard:str, is_tomo:bool)->None:
        """
        :param image_dir path to the directory contain the images
        :param box_dir path to the directory contain the annotiation (the boxes)
        :param wildcard string of wildcard
        :param is_tomo True if the case is a tomo False if it is a SPA
        """
        self.box_dir = box_dir
        self.wildcard = wildcard
        self.is_tomo =is_tomo
        self.unsaved_changes = False
        self.current_image_path = None
        self.image_folder_path = image_dir

        # It is True when we are picking a filament. That is a workaround to fix https://gitlab.gwdg.de/mpi-dortmund/sphire/cryolo-boxmanager/-/issues/50
        # it seems that releasing the button is faster than changing the "event.button" internal var that causes the bug
        self.picking_filament = False

        # the filament we are picking in real time
        self.current_pick_filament = None

        # if we are trying to remove a filemant (otherwise onrelease ll create a new one)
        self.removing_picked_filament = False

        self.trace_all = False

        # helper var used for moving an already existing sketch into the image
        self.moving_box = None
        self.is_moving = False      # True when start to move the moving_box. Before setting to True delete the sketch from the 3D particles (in tomo case

        # identifies if the loaded case is SPA, single tomo or folder of tomo
        self.type_case = constants.Type_case.NOT_LOADED.value

        self.use_estimated_size = False

        # tomo var
        self.current_tomoimage_mmap = None        # it contains a numpy array
        self.index_tomo = None
        self.last_index_tomo = None
            # k= tomo_filename. Contains the values for the 'tracing' parameters silder
        self.smallest_image_dim = dict()
        self.tot_frames = dict()

        # in the folder case it keeps the var params (e.g.: confidence threshold)     for each tomo
        self.params = dict()

        # info about the imported boxes:
            # The 'est_size' instance of a Sketch Class contains the estimated size of cryolo
            # when loaded from '.cbox', and should be never changed
        self.est_box_from_cbox = None   # est_size got from cryolo
        self.is_cbox = False            # the data are in cbox format
        self.is_cbox_untraced = False   # the data are in cbox format and represent untraced particles/filaments
        self.has_filament = False       # the data are filament
        self.has_3d = False             # the 3D visualization is available (only cbox tomo cases. when untraced box are not loaded)

        # it ll be a obj 'Box_dictionary' and it ll manage the sketches in all the images and their conversions
        self.box_dictionary = None

        # If True the current boxes will be hidden. Activated pressing 'h'
        # it is still possible to pick. Until we do not press 'h' again only the new picked sketches ll be drawn
        # in case of changing image it will be deactivated
        self.toggle = False

        # list of sketches to visualize. Ideally it contains the rectangles present in the current image/slice.
        # It is basically a reference, hence changing it you change both, of Box_dictionary.box_dictionary var
        # At usual it is fill with Box_dictionary.get_sketches
        self.rectangles = list()

        self.picking_combobox_value = constants.Picking_cb.PARTICLE.value

        # tab visualization vars
        self.visualization = constants.Visualization_cb.RECT.value
        self.boxsize = constants.Default_settings_value.DEFAULT_BOX_SIZE.value
        self.box_distance_filament = constants.Default_settings_value.DEFAULT_BOX_DISTANCE_FILAMENT_SIZE.value

        # tab thresholding vars
        self.current_conf_thresh = constants.Default_settings_value.DEFAULT_CURRENT_CONF_THRESH.value
        self.lower_size_thresh = constants.Default_settings_value.DEFAULT_LOWER_SIZE_THRESH.value
        self.upper_size_thresh = constants.Default_settings_value.DEFAULT_UPPER_SIZE_THRESH.value
        self.current_num_boxes_thresh = constants.Default_settings_value.DEFAULT_CURRENT_NUM_BOXES_THRESH.value
        # value got by cryolo and filled in the import_data step
        self.min_valid_size_thresh = constants.Default_settings_value.DEFAULT_LOWER_SIZE_THRESH.value
        self.max_valid_size_thresh = constants.Default_settings_value.DEFAULT_UPPER_SIZE_THRESH.value

        # tab filtering vars
        self.filter_freq = constants.Default_settings_value.DEFAULT_FILTER_FREQ.value

        # tab tracing vars
        self.search_range =constants.Default_settings_value.DEFAULT_SEARCH_RANGE.value
        self.memory = constants.Default_settings_value.DEFAULT_MEMORY.value
        self.min_length = constants.Default_settings_value.DEFAULT_MIN_LENGTH.value
        self.min_edge_weight = constants.Default_settings_value.DEFAULT_MIN_EDGE_WEIGHT.value
        self.win_size = constants.Default_settings_value.DEFAULT_BOX_SIZE.value

    def set_rectangles(self, rectangles:List[MySketch.MySketch])->None:
        """
        Fill the rectangles var
        :param rectangles: list of Sketches
        """
        self.rectangles = rectangles

    def clean_box_dictionary(self)->None:
        """
        Clean the box_dictionary var.
        It is used only after opening images. Otherwise i could have the entries of the old case
        """
        if isinstance(self.box_dictionary,box_dictionary.Box_dictionary):
            self.box_dictionary.clean()

    def init_box_dictionary(self, onlyfiles:List[str], tot_slices:int = None)->None:
        """
        Init the box_dictionary variable
        :param onlyfiles: list of the files present in the folder, single file (into a var list)in case of TOMO
        :param tot_slices: number of slices present in the TOMO. Only constants.Type_case.TOMO case
        """
        if self.type_case == constants.Type_case.NOT_LOADED.value:
            exit()
        self.box_dictionary = box_dictionary.Box_dictionary(type_case=self.type_case, onlyfiles=onlyfiles, tot_slices=tot_slices)

    def set_tomo_vars(self)->None:
        """
        Set the tomo vars after opening a tomo or a folder of tomos
        """
        self.index_tomo =0
        self.last_index_tomo = 0
        self.current_tomoimage_mmap = None  # i reset it when i open a new image/folder of
        self.smallest_image_dim = dict()
        self.tot_frames = dict()


    def get_filter_state(self)->tuple:
        return self.lower_size_thresh,self.upper_size_thresh,self.current_conf_thresh

    def get_threshold_params(self, filename:str = None)->tuple:
        """
        Return the param for the current image
        :param filename: in folder case i have to specify from which file, if None returns those of the current case
        :return the 4 threshold settable from the visualization's tab
        """
        if self.type_case == constants.Type_case.TOMO_FOLDER.value and filename:
            p = self.params[filename]
            return p.conf_thresh, p.upper_size_thresh, p.lower_size_thresh, p.num_boxes_thresh
        return self.current_conf_thresh, self.upper_size_thresh, self.lower_size_thresh, self.current_num_boxes_thresh


    def fill_params(self,list_files:List[str])->None:
        """
        filter_freq, upper_size_thresh, lower_size_thresh, conf_thresh, min_num_boxes_thresh,
                 max_num_boxes_thresh, min_edge_weight,
                 search_range, memory, min_length
        """
        for f in list_files:
            self.params.update({f: helper.Params(f, constants.Default_settings_value.DEFAULT_FILTER_FREQ.value, constants.Default_settings_value.DEFAULT_BOX_SIZE.value, constants.Default_settings_value.DEFAULT_UPPER_SIZE_THRESH.value, constants.Default_settings_value.DEFAULT_LOWER_SIZE_THRESH.value,
                                          constants.Default_settings_value.DEFAULT_CURRENT_CONF_THRESH.value, constants.Default_settings_value.DEFAULT_CURRENT_NUM_BOXES_THRESH.value,
                                          constants.Default_settings_value.DEFAULT_MIN_EDGE_WEIGHT.value, constants.Default_settings_value.DEFAULT_SEARCH_RANGE.value,
                                          constants.Default_settings_value.DEFAULT_MEMORY.value, constants.Default_settings_value.DEFAULT_MIN_LENGTH.value, constants.Default_settings_value.DEFAULT_BOX_SIZE.value)})

    def reset_config(self)->None:
        """
        Restore the initial values of the model. It is called after clicking 'File-Reset'
        """
        self.unsaved_changes = False
        self.rectangles = list()
        self.moving_box = None
        self.is_moving = False

        self.current_pick_filament = None
        self.removing_picked_filament = False
        self.picking_filament = False

        # In case of loading new images it ll be overwritten in the open function (e.g.:open_tomo_folder)
        if self.box_dictionary:
            self.box_dictionary.reset()

        # The following three parameters are filled via cml hence when we reset they ll be set to None
        self.box_dir = None
        self.wildcard = None
        self.is_tomo = None

        self.use_estimated_size = False

        # info about the imported boxes:
        self.is_cbox = False
        self.is_cbox_untraced = False
        self.has_filament = False
        self.est_box_from_cbox = None

        self.toggle = False

        self.picking_combobox_value = constants.Picking_cb.PARTICLE.value

        # tab visualization vars
        self.visualization = constants.Visualization_cb.RECT.value
        self.boxsize = constants.Default_settings_value.DEFAULT_BOX_SIZE.value
        self.box_distance_filament = constants.Default_settings_value.DEFAULT_BOX_DISTANCE_FILAMENT_SIZE.value

        # tab thresholding vars
        self.current_conf_thresh = constants.Default_settings_value.DEFAULT_CURRENT_CONF_THRESH.value
        self.lower_size_thresh = constants.Default_settings_value.DEFAULT_LOWER_SIZE_THRESH.value
        self.upper_size_thresh = constants.Default_settings_value.DEFAULT_UPPER_SIZE_THRESH.value
        self.current_num_boxes_thresh = constants.Default_settings_value.DEFAULT_CURRENT_NUM_BOXES_THRESH.value
        # value got by cryolo and filled in the import_data step
        self.min_valid_size_thresh = constants.Default_settings_value.DEFAULT_LOWER_SIZE_THRESH.value
        self.max_valid_size_thresh = constants.Default_settings_value.DEFAULT_UPPER_SIZE_THRESH.value

        # tab filtering vars
        self.filter_freq = constants.Default_settings_value.DEFAULT_FILTER_FREQ.value

        # tab tracing vars
        self.search_range =constants.Default_settings_value.DEFAULT_SEARCH_RANGE.value
        self.memory = constants.Default_settings_value.DEFAULT_MEMORY.value
        self.min_length = constants.Default_settings_value.DEFAULT_MIN_LENGTH.value
        self.min_edge_weight = constants.Default_settings_value.DEFAULT_MIN_EDGE_WEIGHT.value
        self.win_size = constants.Default_settings_value.DEFAULT_BOX_SIZE.value

        self.trace_all = False

        # The following vars will be not reset:
        # self.type_case
        #   no reset the next 2 vars, otherwise after resetting we cannot change image
        #self.current_image_path = None
        #self.image_folder_path = None
        # no reset the tomo var too , otherwise after resetting we cannot change image (because it crashes)
        #self.index_3D_tomo = None
        #self.last_index_3D_tomo = None
        #self.current_tomoimage_mmap = None

        # self.has_3d = False